import re

# css: wrapper1{wrapper2{selector{styles}}}
# wrapper: @xxx aaa and (ccc: ddd)
# selector: parent-modifier1 selector[attr=value]:modifier2
# styles: key1:value1;key2:value2;
append_template = '{selector}:{modifier}'
append_template2 = '{selector}::{modifier}'
selector_modifier_templates = {
    # pseudo classes
    'hover':             append_template,              # &:hover
    'focus':             append_template,              # &:focus
    'focus-within':      append_template,              # &:focus-within
    'focus-visible':     append_template,              # &:focus-visible
    'active':            append_template,              # &:active
    'visited':           append_template,              # &:visited
    'target':            append_template,              # &:target
    'first':             '{selector}:first-child',     # &:first-child
    'last':              '{selector}:last-child',      # &:last-child
    'only':              '{selector}:only-child',      # &:only-child
    'odd':               '{selector}:nth-child(odd)',  # &:nth-child(odd)
    'even':              '{selector}:nth-child(even)', # &:nth-child(even)
    'first-of-type':     append_template,              # &:first-of-type
    'last-of-type':      append_template,              # &:last-of-type
    'only-of-type':      append_template,              # &:only-of-type
    'empty':             append_template,              # &:empty
    'disabled':          append_template,              # &:disabled
    'enabled':           append_template,              # &:enabled
    'checked':           append_template,              # &:checked
    'indeterminate':     append_template,              # &:indeterminate
    'default':           append_template,              # &:default
    'required':          append_template,              # &:required
    'valid':             append_template,              # &:valid
    'invalid':           append_template,              # &:invalid
    'in-range':          append_template,              # &:in-range
    'out-of-range':      append_template,              # &:out-of-range
    'placeholder-shown': append_template,              # &:placeholder-shown
    'autofill':          append_template,              # &:autofill
    'read-only':         append_template,              # &:read-only

    'before':            append_template2,             # &::before
    'after':             append_template2,             # &::after
    'first-letter':      append_template2,             # &::first-letter
    'first-line':        append_template2,             # &::first-line
    'marker':            append_template2,             # &::marker
    'selection':         append_template2,             # &::selection
    'file':              append_template2,             # &::file-selector-button
    'backdrop':          append_template2,             # &::backdrop
    'placeholder':       append_template2,             # &::placeholder


    # aria
    'aria-checked':      '{selector}[aria-checked="true"]',  # &[aria-checked="true"]
    'aria-disabled':     '{selector}[aria-disabled="true"]', # &[aria-disabled="true"]
    'aria-expanded':     '{selector}[aria-expanded="true"]', # &[aria-expanded="true"]
    'aria-hidden':       '{selector}[aria-hidden="true"]',   # &[aria-hidden="true"]
    'aria-pressed':      '{selector}[aria-pressed="true"]',  # &[aria-pressed="true"]
    'aria-readonly':     '{selector}[aria-readonly="true"]', # &[aria-readonly="true"]
    'aria-required':     '{selector}[aria-required="true"]', # &[aria-required="true"]
    'aria-selected':     '{selector}[aria-selected="true"]', # &[aria-selected="true"]

    'rtl':                '[dir="rtl"] {selector}',          # [dir="rtl"] &
    'ltr':                '[dir="ltr"] {selector}',          # [dir="ltr"] &

    'open':               '{selector}[open]',                # &[open]

    # dark mode
    'dark':               '.dark {selector}, [dark=""] {selector}', # .dark & 
}

selector_modifiers = set(selector_modifier_templates.keys())

# 原先modifier是有中括号的，所以命名为re_brackets，现在把中括号去掉了。
#另外，为防止将utility误判为modifier，像"max-w-sm"之类的utility不使用re_brackets
#最后，发现很多正则没那么多共同点，不需要定义一个公共的正则了，去掉
#re_brackets = '(.+)'
re_selector_modifier_templates = {
    f'aria-(.+)':          r'{selector}[aria-{group1}]',       # &[aria-xxx]
    f'data-(.+)':          r'{selector}[data-{group1}]',       # &[data-xxx]
    f'nth-(\d+)':          r'{selector}:nth-child({group1})',  # &:nth-child(ddd)
    f'group-([^@]+)':      r'.group:{group1} {selector}, [group=""]:{group1} {selector}',
    f'group-([^@]+)@(.+)': r'.group\@{group2}:{group1} {selector}, [group\@{group2}=""]:{group1} {selector}',
    f'peer-([^@]+)':       r'.peer:{group1} ~ {selector}, [peer=""]:{group1} ~ {selector}',
    f'peer-([^@]+)@(.+)':  r'.peer\@{group2}:{group1} + {selector}, [peer\@{group2}=""]:{group1} ~ {selector}',
}
re_selector_modifiers = set(re_selector_modifier_templates.keys())

wrapper_modifier_templates = {
    # responsive design
    'sm':            ('@media (min-width: 640px)', 640),                # @media (min-width: 640px)  { ... }
    'md':            ('@media (min-width: 768px)', 768),                # @media (min-width: 768px)  { ... }
    'lg':            ('@media (min-width: 1024px)', 1024),              # @media (min-width: 1024px) { ... }
    'xl':            ('@media (min-width: 1280px)', 1280),              # @media (min-width: 1280px) { ... }
    '2xl':           ('@media (min-width: 1536px)', 1536),              # @media (min-width: 1536px) { ... }
    'max-sm':        ('@media not all and (min-width: 640px)', -640),   # @media not all and (min-width: 640px) { ... }
    'max-md':        ('@media not all and (min-width: 768px)', -768),   # @media not all and (min-width: 768px) { ... }
    'max-lg':        ('@media not all and (min-width: 1024px)', -1024), # @media not all and (min-width: 1024px) { ... }
    'max-xl':        ('@media not all and (min-width: 1280px)', -1280), # @media not all and (min-width: 1280px) { ... }
    'max-2xl':       ('@media not all and (min-width: 1536px)', -1536), # @media not all and (min-width: 1536px) { ... }

    #'dark':          '@media (perfers-color-scheme: dark)',             # @media (perfers-color-scheme: dark) { ... }

    'portrait':      '@media (orientation: portrait)',                  # @media (orientation: portrait) { ... }
    'landscape':     '@media (orientation: landscape)',                 # @media (orientation: landscape) { ... }
    'motion-safe':   '@media (prefers-reduced-motion: no-preference)',  # @media (prefers-reduced-motion: no-preference) { ... }
    'motion-reduce': '@media (prefers-reduced-motion: reduce)',         # @media (prefers-reduced-motion: reduce) { ... }
    'contrast-more': '@media (prefers-contrast: more)',                 # @media (prefers-contrast: more) { ... }
    'contrast-less': '@media (prefers-contrast: less)',                 # @media (prefers-contrast: less) { ... }
    'print':         '@media print',                                    # @media print { ... }

    # add for support plugin
    'hover-hover':   '@media (hover: hover)',                           # @media (hover: hover) { ... }
    'hover-none':    '@media (hover: none)',                            # @media (hover: none) { ... }
}
wrapper_modifiers = set(wrapper_modifier_templates.keys())

re_wrapper_modifier_templates = {
    f'min-([0-9]+)px': ('@media (min-width: {group1}px)', '{group1}'),  # @media (min-width: xxx) { ... }
    f'max-([0-9]+)px': ('@media (max-width: {group1}px)', '-{group1}'), # @media (max-width: xxx) { ... }
    f'supports-(.+)':  '@supports ({group1})',                          # @supports (xxx) { ... }
}
re_wrapper_modifiers = set(re_wrapper_modifier_templates.keys())

all_modifiers = set(selector_modifiers)
all_modifiers.update(wrapper_modifiers)

re_all_modifiers = set(re_selector_modifiers)
re_all_modifiers.update(re_wrapper_modifiers)

def is_modifier(value):
    return (value in all_modifiers or
            any(re.fullmatch(pattern, value) for pattern in re_all_modifiers))


def is_modifiers(value):
    return value and all(is_modifier(v) for v in value.split(':'))


def add_modifier(css, modifier):
    def convert_template(template):
        if isinstance(template, tuple):
            template, screen_order = template
            css.screen_order.append(screen_order)
        return template

    def convert_re_template(template, group1, group2):
        if isinstance(template, tuple):
            template, screen_order = template
            css.screen_order.append(screen_order)
        return template

    if modifier in selector_modifiers:
        template = convert_template(selector_modifier_templates[modifier])
        css.selector_template = template.format(selector=css.selector_template, modifier=modifier)
        return True
    elif modifier in wrapper_modifiers:
        template = convert_template(wrapper_modifier_templates[modifier])
        css.wrappers.append(template)
        return True
    else:
        for pattern in re_selector_modifiers:
            match = re.fullmatch(pattern, modifier)
            if match:
                group1 = group2 = None
                ngroup = len(match.groups())
                if ngroup > 0:
                    group1 = css.quote(match.group(1))
                if ngroup > 1:
                    group2 = css.quote(match.group(2))
                template = convert_re_template(re_selector_modifier_templates[pattern], group1, group2)
                css.selector_template = template.format(selector=css.selector_template, group1=group1, group2=group2)
                return True

        for pattern in re_wrapper_modifiers:
            match = re.fullmatch(pattern, modifier)
            if match:
                group1 = group2 = None
                ngroup = len(match.groups())
                if ngroup > 0:
                    group1 = css.quote(match.group(1))
                if ngroup > 1:
                    group2 = css.quote(match.group(2))
                template = convert_re_template(re_wrapper_modifier_templates[pattern], group1, group2)
                css.wrappers.append(template.format(group1=group1, group2=group2))
                return True
        return False



from django.core.management.base import CommandError, LabelCommand
from fryhcs.utils import create_js_generator, create_css_generator
from fryhcs.fry.generator import fry_to_py
from pathlib import Path


class Command(LabelCommand):
    help = "Runs fryhcs commands"
    missing_args_message = """
Command argument is missing, please add one of the following:
  build - to compile .fry into production css and js
  topy - to compile .fry into .py file
Usage example:
  python manage.py fryhcs build
  python manage.py fryhcs topy FYFILE
"""

    def handle(self, *labels, **options):
        if len(labels) == 1 and labels[0] == 'build':
            return self.build()
        elif len(labels) == 2 and labels[0] == 'topy':
            return self.topy(labels[1])
        else:
            return "Wrong command, Usage: python manage.py fryhcs [build | topy FYFILE]"

    def build(self)
        output = []
        js_generator = create_js_generator()
        css_generator = create_css_generator()
        output.append("Processing css information in the following place:")
        output.append('')
        for file in css_generator.input_files:
            output.append(f"  {file}")
        output.append('')
        css_generator.generate()
        output.append("... Done.")
        output.append('')
        output.append(f"CSS file {css_generator.output_file} is regenerated.")
        output.append('')
        output.append("Processing js information in the following place:")
        output.append('')
        for file in js_generator.fileiter.all_files():
            output.append(f"  {file}")
        output.append('')
        js_generator.generate()
        output.append("... Done.")
        output.append('')
        output.append('')
        return '\n'.join(output)

    def topy(fryfile):
        path = Path(fryfile)
        if not path.is_file():
            return f"Wrong argument to fryhcs topy command: {fryfile} is not readable"
        with path.open('r', encoding='utf-8') as f:
            data = f.read()
        return fry_to_py(data, path)



"""
    fryhcs.cmdline
    ~~~~~~~~~~~~~~~~

    Command line interface.

    :copyright: Copyright 2023 by zenkj<juzejian@gmail.com>
    :license: BSD, see LICENSE for details.
"""


import fnmatch
import os
import subprocess
import sys
import threading
import time
import traceback
import typing as t
from itertools import chain
from pathlib import PurePath, Path

import werkzeug
from werkzeug.serving import make_server
from flask.cli import FlaskGroup, shell_command, routes_command, CertParamType, pass_script_info, get_debug_flag, _debug_option

import click

from fryhcs.utils import create_css_generator, create_js_generator

css_generator = None 
js_generator = None

import logging

logger = logging.getLogger(__name__)

_log_add_style = True

if os.name == "nt":
    try:
        __import__("colorama")
    except ImportError:
        _log_add_style = False


# The various system prefixes where imports are found. Base values are
# different when running in a virtualenv. All reloaders will ignore the
# base paths (usually the system installation). The stat reloader won't
# scan the virtualenv paths, it will only include modules that are
# already imported.
_ignore_always = tuple({sys.base_prefix, sys.base_exec_prefix})
prefix = {*_ignore_always, sys.prefix, sys.exec_prefix}

if hasattr(sys, "real_prefix"):
    # virtualenv < 20
    prefix.add(sys.real_prefix)

_stat_ignore_scan = tuple(prefix)
del prefix
_ignore_common_dirs = {
    "__pycache__",
    ".git",
    ".hg",
    ".tox",
    ".nox",
    ".pytest_cache",
    ".mypy_cache",
}


def _iter_module_paths() -> t.Iterator[str]:
    """Find the filesystem paths associated with imported modules."""
    # List is in case the value is modified by the app while updating.
    for module in list(sys.modules.values()):
        name = getattr(module, "__file__", None)

        if name is None or name.startswith(_ignore_always):
            continue

        while not os.path.isfile(name):
            # Zip file, find the base file without the module path.
            old = name
            name = os.path.dirname(name)

            if name == old:  # skip if it was all directories somehow
                break
        else:
            yield name


def _remove_by_pattern(paths, exclude_patterns):
    for pattern in exclude_patterns:
        paths.difference_update(fnmatch.filter(paths, pattern))


def _find_stat_paths(extra_files, exclude_patterns) -> t.Iterable[str]:
    """Find paths for the stat reloader to watch. Returns imported
    module files, Python files under non-system paths. Extra files and
    Python files under extra directories can also be scanned.

    System paths have to be excluded for efficiency. Non-system paths,
    such as a project root or ``sys.path.insert``, should be the paths
    of interest to the user anyway.
    """
    paths = set()

    for path in chain(list(sys.path), extra_files):
        path = os.path.abspath(path)

        if os.path.isfile(path):
            # zip file on sys.path, or extra file
            paths.add(path)
            continue

        parent_has_py = {os.path.dirname(path): True}

        for root, dirs, files in os.walk(path):
            # Optimizations: ignore system prefixes, __pycache__ will
            # have a py or pyc module at the import path, ignore some
            # common known dirs such as version control and tool caches.
            if (
                root.startswith(_stat_ignore_scan)
                or os.path.basename(root) in _ignore_common_dirs
            ):
                dirs.clear()
                continue

            has_py = False

            for name in files:
                if name.endswith((".py", ".pyc")):
                    has_py = True
                    paths.add(os.path.join(root, name))

            # Optimization: stop scanning a directory if neither it nor
            # its parent contained Python files.
            if not (has_py or parent_has_py[os.path.dirname(root)]):
                dirs.clear()
                continue

            parent_has_py[root] = has_py

    paths.update(_iter_module_paths())
    _remove_by_pattern(paths, exclude_patterns)
    return paths


def _find_watchdog_paths(extra_files, exclude_patterns) -> t.Iterable[str]:
    """Find paths for the stat reloader to watch. Looks at the same
    sources as the stat reloader, but watches everything under
    directories instead of individual files.
    """
    dirs = set()

    for name in chain(list(sys.path), extra_files):
        name = os.path.abspath(name)

        if os.path.isfile(name):
            name = os.path.dirname(name)

        dirs.add(name)

    for name in _iter_module_paths():
        dirs.add(os.path.dirname(name))

    _remove_by_pattern(dirs, exclude_patterns)
    return _find_common_roots(dirs)


def _find_common_roots(paths: t.Iterable[str]) -> t.Iterable[str]:
    root: dict[str, dict] = {}

    for chunks in sorted((PurePath(x).parts for x in paths), key=len, reverse=True):
        node = root

        for chunk in chunks:
            node = node.setdefault(chunk, {})

        node.clear()

    rv = set()

    def _walk(node: t.Mapping[str, dict], path: tuple[str, ...]) -> None:
        for prefix, child in node.items():
            _walk(child, path + (prefix,))

        if not node:
            rv.add(os.path.join(*path))

    _walk(root, ())
    return rv


def _get_args_for_reloading():
    """Determine how the script was executed, and return the args needed
    to execute it again in a new process.
    """
    if sys.version_info >= (3, 10):
        # sys.orig_argv, added in Python 3.10, contains the exact args used to invoke
        # Python. Still replace argv[0] with sys.executable for accuracy.
        return [sys.executable, *sys.orig_argv[1:]]

    rv = [sys.executable]
    py_script = sys.argv[0]
    args = sys.argv[1:]
    # Need to look at main module to determine how it was executed.
    __main__ = sys.modules["__main__"]

    # The value of __package__ indicates how Python was called. It may
    # not exist if a setuptools script is installed as an egg. It may be
    # set incorrectly for entry points created with pip on Windows.
    if getattr(__main__, "__package__", None) is None or (
        os.name == "nt"
        and __main__.__package__ == ""
        and not os.path.exists(py_script)
        and os.path.exists(f"{py_script}.exe")
    ):
        # Executed a file, like "python app.py".
        py_script = os.path.abspath(py_script)

        if os.name == "nt":
            # Windows entry points have ".exe" extension and should be
            # called directly.
            if not os.path.exists(py_script) and os.path.exists(f"{py_script}.exe"):
                py_script += ".exe"

            if (
                os.path.splitext(sys.executable)[1] == ".exe"
                and os.path.splitext(py_script)[1] == ".exe"
            ):
                rv.pop(0)

        rv.append(py_script)
    else:
        # Executed a module, like "python -m werkzeug.serving".
        if os.path.isfile(py_script):
            # Rewritten by Python from "-m script" to "/path/to/script.py".
            py_module = t.cast(str, __main__.__package__)
            name = os.path.splitext(os.path.basename(py_script))[0]

            if name != "__main__":
                py_module += f".{name}"
        else:
            # Incorrectly rewritten by pydevd debugger from "-m script" to "script".
            py_module = py_script

        rv.extend(("-m", py_module.lstrip(".")))

    rv.extend(args)
    return rv


class ReloaderLoop:
    name = ""

    def __init__(
        self,
        extra_files = None,
        exclude_patterns = None,
        interval = 1,
    ) -> None:
        self.extra_files: set[str] = {os.path.abspath(x) for x in extra_files or ()}
        self.exclude_patterns: set[str] = set(exclude_patterns or ())
        self.interval = interval

    def __enter__(self):
        """Do any setup, then run one step of the watch to populate the
        initial filesystem state.
        """
        self.run_step()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        """Clean up any resources associated with the reloader."""
        pass

    def run(self) -> None:
        """Continually run the watch step, sleeping for the configured
        interval after each step.
        """
        while True:
            self.run_step()
            time.sleep(self.interval)

    def run_step(self) -> None:
        """Run one step for watching the filesystem. Called once to set
        up initial state, then repeatedly to update it.
        """
        pass

    def restart_with_reloader(self) -> int:
        """Spawn a new Python interpreter with the same arguments as the
        current one, but running the reloader thread.
        """
        while True:
            logger.info(f" * Restarting with {self.name}")
            args = _get_args_for_reloading()
            new_environ = os.environ.copy()
            new_environ["FRYHCS_RUN_MAIN"] = "true"
            exit_code = subprocess.call(args, env=new_environ, close_fds=False)

            if exit_code != 3:
                return exit_code

    def trigger_reload(self, filename: str) -> None:
        self.log_reload(filename)
        generate_static(filename)
        sys.exit(3)

    def log_reload(self, filename: str) -> None:
        filename = os.path.abspath(filename)
        logger.info(f" * Detected change in {filename!r}, reloading")


class StatReloaderLoop(ReloaderLoop):
    name = "stat"

    def __enter__(self) -> ReloaderLoop:
        self.mtimes: dict[str, float] = {}
        return super().__enter__()

    def run_step(self) -> None:
        for name in _find_stat_paths(self.extra_files, self.exclude_patterns):
            try:
                mtime = os.stat(name).st_mtime
            except OSError:
                continue

            old_time = self.mtimes.get(name)

            if old_time is None:
                self.mtimes[name] = mtime
                continue

            if mtime > old_time:
                self.trigger_reload(name)


class WatchdogReloaderLoop(ReloaderLoop):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        from watchdog.observers import Observer
        from watchdog.events import PatternMatchingEventHandler
        from watchdog.events import EVENT_TYPE_OPENED
        from watchdog.events import FileModifiedEvent

        super().__init__(*args, **kwargs)
        trigger_reload = self.trigger_reload

        class EventHandler(PatternMatchingEventHandler):
            def on_any_event(self, event: FileModifiedEvent):  # type: ignore
                if event.event_type == EVENT_TYPE_OPENED:
                    return

                trigger_reload(event.src_path)

        reloader_name = Observer.__name__.lower()  # type: ignore[attr-defined]

        if reloader_name.endswith("observer"):
            reloader_name = reloader_name[:-8]

        self.name = f"watchdog ({reloader_name})"
        self.observer = Observer()
        # Extra patterns can be non-Python files, match them in addition
        # to all Python files in default and extra directories. Ignore
        # __pycache__ since a change there will always have a change to
        # the source file (or initial pyc file) as well. Ignore Git and
        # Mercurial internal changes.
        extra_patterns = [p for p in self.extra_files if not os.path.isdir(p)]
        self.event_handler = EventHandler(
            patterns=["*.py", "*.pyc", "*.zip", *extra_patterns],
            ignore_patterns=[
                *[f"*/{d}/*" for d in _ignore_common_dirs],
                *self.exclude_patterns,
            ],
        )
        self.should_reload = False

    def trigger_reload(self, filename: str) -> None:
        # This is called inside an event handler, which means throwing
        # SystemExit has no effect.
        # https://github.com/gorakhargosh/watchdog/issues/294
        self.should_reload = True
        self.log_reload(filename)
        generate_static(filename)


    def __enter__(self) -> ReloaderLoop:
        self.watches: dict[str, t.Any] = {}
        self.observer.start()
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        self.observer.stop()
        self.observer.join()

    def run(self) -> None:
        while not self.should_reload:
            self.run_step()
            time.sleep(self.interval)

        sys.exit(3)

    def run_step(self) -> None:
        to_delete = set(self.watches)

        for path in _find_watchdog_paths(self.extra_files, self.exclude_patterns):
            if path not in self.watches:
                try:
                    self.watches[path] = self.observer.schedule(
                        self.event_handler, path, recursive=True
                    )
                except OSError:
                    # Clear this path from list of watches We don't want
                    # the same error message showing again in the next
                    # iteration.
                    self.watches[path] = None

            to_delete.discard(path)

        for path in to_delete:
            watch = self.watches.pop(path, None)

            if watch is not None:
                self.observer.unschedule(watch)


reloader_loops = {
    "stat": StatReloaderLoop,
    "watchdog": WatchdogReloaderLoop,
}

try:
    __import__("watchdog.observers")
except ImportError:
    reloader_loops["auto"] = reloader_loops["stat"]
else:
    reloader_loops["auto"] = reloader_loops["watchdog"]


def ensure_echo_on() -> None:
    """Ensure that echo mode is enabled. Some tools such as PDB disable
    it which causes usability issues after a reload."""
    # tcgetattr will fail if stdin isn't a tty
    if sys.stdin is None or not sys.stdin.isatty():
        return

    try:
        import termios
    except ImportError:
        return

    attributes = termios.tcgetattr(sys.stdin)

    if not attributes[3] & termios.ECHO:
        attributes[3] |= termios.ECHO
        termios.tcsetattr(sys.stdin, termios.TCSANOW, attributes)


def run_with_reloader(
    main_func,
    extra_files = None,
    exclude_patterns = None,
    interval = 1,
    reloader_type = "auto",
) -> None:
    """Run the given function in an independent Python interpreter."""
    import signal

    signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))
    reloader = reloader_loops[reloader_type](
        extra_files=extra_files, exclude_patterns=exclude_patterns, interval=interval
    )

    try:
        if os.environ.get("FRYHCS_RUN_MAIN") == "true":
            ensure_echo_on()
            t = threading.Thread(target=main_func, args=())
            t.daemon = True

            # Enter the reloader to set up initial state, then start
            # the app thread and reloader update loop.
            with reloader:
                t.start()
                reloader.run()
        else:
            sys.exit(reloader.restart_with_reloader())
    except KeyboardInterrupt:
        pass


def is_running_from_reloader() -> bool:
    """Check if the server is running as a subprocess within the
    fryhcs reloader.

    .. versionadded:: 0.10
    """
    return os.environ.get("FRYHCS_RUN_MAIN") == "true"

def _ansi_style(value: str, *styles: str) -> str:
    if not _log_add_style:
        return value

    codes = {
        "bold": 1,
        "red": 31,
        "green": 32,
        "yellow": 33,
        "magenta": 35,
        "cyan": 36,
    }

    for style in styles:
        value = f"\x1b[{codes[style]}m{value}"

    return f"{value}\x1b[0m"


def run(
    hostname,
    port,
    app,
    use_reloader = True,
    use_debugger = True,
    use_evalex = True,
    extra_files = None,
    exclude_patterns = None,
    reloader_interval = 0.5,
    reloader_type = "auto",
    threaded = False,
    processes = 1,
    request_handler = None,
    static_files = {},
    passthrough_errors = False,
    ssl_context = None,
) -> None:
    """Start a development server for a WSGI application. Various
    optional features can be enabled.

    .. warning::

        Do not use the development server when deploying to production.
        It is intended for use only during local development. It is not
        designed to be particularly efficient, stable, or secure.

    :param hostname: The host to bind to, for example ``'localhost'``.
        Can be a domain, IPv4 or IPv6 address, or file path starting
        with ``unix://`` for a Unix socket.
    :param port: The port to bind to, for example ``8080``. Using ``0``
        tells the OS to pick a random free port.
    :param application: The WSGI application to run.
    :param use_reloader: Use a reloader process to restart the server
        process when files are changed.
    :param use_debugger: Use Werkzeug's debugger, which will show
        formatted tracebacks on unhandled exceptions.
    :param use_evalex: Make the debugger interactive. A Python terminal
        can be opened for any frame in the traceback. Some protection is
        provided by requiring a PIN, but this should never be enabled
        on a publicly visible server.
    :param extra_files: The reloader will watch these files for changes
        in addition to Python modules. For example, watch a
        configuration file.
    :param exclude_patterns: The reloader will ignore changes to any
        files matching these :mod:`fnmatch` patterns. For example,
        ignore cache files.
    :param reloader_interval: How often the reloader tries to check for
        changes.
    :param reloader_type: The reloader to use. The ``'stat'`` reloader
        is built in, but may require significant CPU to watch files. The
        ``'watchdog'`` reloader is much more efficient but requires
        installing the ``watchdog`` package first.
    :param threaded: Handle concurrent requests using threads. Cannot be
        used with ``processes``.
    :param processes: Handle concurrent requests using up to this number
        of processes. Cannot be used with ``threaded``.
    :param request_handler: Use a different
        :class:`~BaseHTTPServer.BaseHTTPRequestHandler` subclass to
        handle requests.
    :param static_files: A dict mapping URL prefixes to directories to
        serve static files from using
        :class:`~werkzeug.middleware.SharedDataMiddleware`.
    :param passthrough_errors: Don't catch unhandled exceptions at the
        server level, let the server crash instead. If ``use_debugger``
        is enabled, the debugger will still catch such errors.
    :param ssl_context: Configure TLS to serve over HTTPS. Can be an
        :class:`ssl.SSLContext` object, a ``(cert_file, key_file)``
        tuple to create a typical context, or the string ``'adhoc'`` to
        generate a temporary self-signed certificate.

    .. versionchanged:: 2.1
        Instructions are shown for dealing with an "address already in
        use" error.

    .. versionchanged:: 2.1
        Running on ``0.0.0.0`` or ``::`` shows the loopback IP in
        addition to a real IP.

    .. versionchanged:: 2.1
        The command-line interface was removed.

    .. versionchanged:: 2.0
        Running on ``0.0.0.0`` or ``::`` shows a real IP address that
        was bound as well as a warning not to run the development server
        in production.

    .. versionchanged:: 2.0
        The ``exclude_patterns`` parameter was added.

    .. versionchanged:: 0.15
        Bind to a Unix socket by passing a ``hostname`` that starts with
        ``unix://``.

    .. versionchanged:: 0.10
        Improved the reloader and added support for changing the backend
        through the ``reloader_type`` parameter.

    .. versionchanged:: 0.9
        A command-line interface was added.

    .. versionchanged:: 0.8
        ``ssl_context`` can be a tuple of paths to the certificate and
        private key files.

    .. versionchanged:: 0.6
        The ``ssl_context`` parameter was added.

    .. versionchanged:: 0.5
       The ``static_files`` and ``passthrough_errors`` parameters were
       added.
    """
    if not isinstance(port, int):
        raise TypeError("port must be an integer")

    global css_generator, js_generator
    with app.app_context():
        css_generator = create_css_generator()
        js_generator = create_js_generator()

    if use_reloader:
        from fryhcs.reload import event_stream, mime_type
        from flask import Response

        @app.get('/_check_hotreload')
        def fryhcs_check_hotreload():
            return Response(event_stream(), mimetype=mime_type)

    static_files = static_files if static_files else {}

    with app.app_context():
        from fryhcs.config import fryconfig
        static_files[fryconfig.static_url] = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'static')
    
    if static_files:
        from werkzeug.middleware.shared_data import SharedDataMiddleware

        app.wsgi_app = SharedDataMiddleware(app.wsgi_app, static_files)

    if use_debugger:
        from werkzeug.debug import DebuggedApplication

        app.wsgi_app = DebuggedApplication(app.wsgi_app, evalex=use_evalex)

    if not is_running_from_reloader():
        fd = None
    else:
        fd = int(os.environ["FRYHCS_SERVER_FD"])

    srv = make_server(
        hostname,
        port,
        app,
        threaded,
        processes,
        request_handler,
        passthrough_errors,
        ssl_context,
        fd=fd,
    )
    srv.socket.set_inheritable(True)
    os.environ["FRYHCS_SERVER_FD"] = str(srv.fileno())

    if not is_running_from_reloader():
        srv.log_startup()
        with app.app_context():
            generate_static()
        logger.info(_ansi_style("Press CTRL+C to quit", "yellow"))

    if use_reloader:
        try:
            run_with_reloader(
                srv.serve_forever,
                extra_files=extra_files,
                exclude_patterns=exclude_patterns,
                interval=reloader_interval,
                reloader_type=reloader_type,
            )
        finally:
            srv.server_close()
    else:
        srv.serve_forever()


def generate_static(path=None):
    if not css_generator or not js_generator:
        return
    if not path:
        logger.info("Regenerating all js files and css files from .fry files...")
        css_generator.generate()
        js_generator.generate(clean=True)
    else:
        path = Path(path).absolute()
        if path.suffix == '.fry':
            logger.info(f"Regenerating js files and css file for {str(path)}...")
            css_generator.generate(path)
            js_generator.generate([path])


def show_server_banner(debug, app_import_path):
    """Show extra startup messages the first time the server is run,
    ignoring the reloader.
    """
    if is_running_from_reloader():
        return

    if app_import_path is not None:
        click.echo(f" * Serving Fryhcs app '{app_import_path}'")

    if debug is not None:
        click.echo(f" * Debug mode: {'on' if debug else 'off'}")


def _validate_key(ctx, param, value):
    """The ``--key`` option must be specified when ``--cert`` is a file.
    Modifies the ``cert`` param to be a ``(cert, key)`` pair if needed.
    """
    cert = ctx.params.get("cert")
    is_adhoc = cert == "adhoc"

    try:
        import ssl
    except ImportError:
        is_context = False
    else:
        is_context = isinstance(cert, ssl.SSLContext)

    if value is not None:
        if is_adhoc:
            raise click.BadParameter(
                'When "--cert" is "adhoc", "--key" is not used.', ctx, param
            )

        if is_context:
            raise click.BadParameter(
                'When "--cert" is an SSLContext object, "--key is not used.', ctx, param
            )

        if not cert:
            raise click.BadParameter('"--cert" must also be specified.', ctx, param)

        ctx.params["cert"] = cert, value

    else:
        if cert and not (is_adhoc or is_context):
            raise click.BadParameter('Required when using "--cert".', ctx, param)

    return value


class SeparatedPathType(click.Path):
    """Click option type that accepts a list of values separated by the
    OS's path separator (``:``, ``;`` on Windows). Each value is
    validated as a :class:`click.Path` type.
    """

    def convert(self, value, param, ctx):
        items = self.split_envvar_value(value)
        super_convert = super().convert
        return [super_convert(item, param, ctx) for item in items]


@click.command("dev", short_help="Run a development server.")
@click.option("--host", "-h", default="127.0.0.1", help="The interface to bind to.")
@click.option("--port", "-p", default=5000, help="The port to bind to.")
@click.option(
    "--cert",
    type=CertParamType(),
    help="Specify a certificate file to use HTTPS.",
    is_eager=True,
)
@click.option(
    "--key",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    callback=_validate_key,
    expose_value=False,
    help="The key file to use when specifying a certificate.",
)
@click.option(
    "--reload/--no-reload",
    default=None,
    help="Enable or disable the reloader. By default the reloader "
    "is active if debug is enabled.",
)
@click.option(
    "--debugger/--no-debugger",
    default=None,
    help="Enable or disable the debugger. By default the debugger "
    "is active if debug is enabled.",
)
@click.option(
    "--with-threads/--without-threads",
    default=True,
    help="Enable or disable multithreading.",
)
@click.option(
    "--extra-files",
    default=None,
    type=SeparatedPathType(),
    help=(
        "Extra files that trigger a reload on change. Multiple paths"
        f" are separated by {os.path.pathsep!r}."
    ),
)
@click.option(
    "--exclude-patterns",
    default=None,
    type=SeparatedPathType(),
    help=(
        "Files matching these fnmatch patterns will not trigger a reload"
        " on change. Multiple patterns are separated by"
        f" {os.path.pathsep!r}."
    ),
)
@pass_script_info
def dev_command(
    info,
    host,
    port,
    reload,
    debugger,
    with_threads,
    cert,
    extra_files,
    exclude_patterns,
):
    """Run a local development server.

    This server is for development purposes only. It does not provide
    the stability, security, or performance of production WSGI servers.

    The reloader and debugger are enabled by default with the '--debug'
    option.
    """

    if 'FLASK_DEBUG' not in os.environ:
        os.environ['FLASK_DEBUG'] = '1'

    try:
        app = info.load_app()
    except Exception as e:
        if is_running_from_reloader():
            # When reloading, print out the error immediately, but raise
            # it later so the debugger or server can handle it.
            traceback.print_exc()
            err = e

            def app(environ, start_response):
                raise err from None

        else:
            # When not reloading, raise the error immediately so the
            # command fails.
            raise e from None

    debug = get_debug_flag()

    if reload is None:
        reload = debug

    if debugger is None:
        debugger = debug

    show_server_banner(debug, info.app_import_path)

    run(
        host,
        port,
        app,
        use_reloader=reload,
        use_debugger=debugger,
        threaded=with_threads,
        ssl_context=cert,
        extra_files=extra_files,
        exclude_patterns=exclude_patterns,
    )

dev_command.params.insert(0, _debug_option)


@click.command("build", short_help="Build js files and css style file for all .fry files.")
@pass_script_info
def build_command(info):
    """Build js files and css style file for all .fry files."""
    app = info.load_app()
    with app.app_context():
        css_generator = create_css_generator()
        js_generator = create_js_generator()
        logger.info("Regenerating all js files and css files from .fry files...")
        css_generator.generate()
        js_generator.generate(clean=True)


@click.command("topy", short_help="Convert specified .fry file into .py file.")
@click.argument("fryfile")
def topy_command(fryfile):
    """Convert specified .fry file into .py file."""
    from fryhcs.fry.generator import fry_to_py
    path = Path(fryfile)
    if not path.is_file():
        print(f"Error: can't open file '{fryfile}'.")
        sys.exit(1)
    with path.open('r', encoding='utf-8') as f:
        data = f.read()
    source = fry_to_py(data, path)
    try:
        from pygments.formatters.terminal import TerminalFormatter
        from pygments.lexers import PythonLexer
        from pygments import highlight
        lexer = PythonLexer()
        fmter = TerminalFormatter(linenos=True)
        click.echo(highlight(source, lexer, fmter))
    except ImportError:
        lines = source.splitlines()
        if lines:
            prefix_len = len(str(len(lines)))
            source = '\n'.join([f'{i:0{prefix_len}}: {line}' for i, line in zip(range(1, len(lines)+1), lines)])
        click.echo(source)


@click.command("tojs", short_help="Convert specified .fry file into .js file(s).")
@click.argument("fryfile")
@click.argument("jsdir")
def tojs_command(fryfile, jsdir):
    """Convert specified .fry file into .js file(s)."""
    from fryhcs.js.generator import JSGenerator
    path = Path(fryfile)
    if not path.is_file():
        print(f"Error: can't open file '{fryfile}'.")
        sys.exit(1)
    generator = JSGenerator([fryfile], jsdir)
    count = generator.generate()
    if count == 0:
        print(f"No js information in '{fryfile}'.")
    else:
        print(f"{count} js files from '{fryfile}' are generated into directory '{jsdir}'")


@click.command("tocss", short_help="Convert specified .fry file into style.css file.")
@click.option("-p", "--plugin", multiple=True, help="Specify a plugin to be loaded")
@click.argument("fryfile")
@click.argument("cssfile")
def tocss_command(plugin, fryfile, cssfile):
    """Convert specified .fry file into style.css file."""
    if plugin:
        sys.path.insert(0, '')
        plugins = ':'.join(plugin)
        os.environ['FRYHCS_PLUGINS'] = plugins
    from fryhcs.css.generator import CSSGenerator
    path = Path(fryfile)
    if not path.is_file():
        print(f"Error: can't open file '{fryfile}'.")
        sys.exit(1)
    generator = CSSGenerator([fryfile], cssfile)
    generator.generate()
    print(f"styles from '{fryfile}' are generated into file '{cssfile}'.")


@click.command("run", short_help="Convert specified .fry file into .py file and execute it.")
@click.option("-m", "module", default=None, help="Specify a module to be run")
@click.argument("fryfile", default=None, required=False, type=click.Path(exists=True, resolve_path=True))
def run_command(module, fryfile):
    """Convert specified .fry file into .py file and execute it."""
    from runpy import run_module
    if module and fryfile:
        click.echo("ONLY ONE of -m MODULE and PYXFILE can be specified.")
        return
    if module:
        sys.path.insert(0, '')
    elif fryfile:
        path = Path(fryfile)
        if path.is_dir():
            sys.path.insert(0, str(path))
            module = '__main__'
        elif path.suffix in ('.py', '.pyc', '.fry'):
            dir = path.parent
            sys.path.insert(0, str(dir))
            module = path.stem
        else:
            click.echo("PYXFILE should be .py, .pyc or .fry file.")
            return
    else:
        click.echo("one of -m MODULE and PYXFILE should be specified.")
        return
    _m = run_module(module, run_name='__main__')


@click.command("hl", short_help="Highlight specified .fry file based on pygments.")
@click.argument("fryfile", default=None, required=False, type=click.Path(exists=True, resolve_path=True))
def hl_command(fryfile):
    """Highlight specified .fry file based on pygments."""
    try:
        from pygments.formatters.terminal import TerminalFormatter
        from pygments import highlight
    except ImportError:
        click.echo("Pygments is not installed, install it via `pip install pygments`")
        return
    from fryhcs.fry.frylexer import FryLexer
    lexer = FryLexer()
    fmter = TerminalFormatter()
    with open(fryfile, 'r', encoding='utf-8') as f:
        source = f.read()
    click.echo(highlight(source, lexer, fmter))


class FryhcsGroup(FlaskGroup):
    def __init__(self, **extra):
        extra.pop('add_default_commands', None)
        super().__init__(add_default_commands=False, **extra) 
        self.add_command(dev_command)
        self.add_command(build_command)
        self.add_command(topy_command)
        self.add_command(tojs_command)
        self.add_command(tocss_command)
        self.add_command(run_command)
        self.add_command(hl_command)
        self.add_command(shell_command)
        self.add_command(routes_command)

cli = FryhcsGroup(
    name="fryhcs",
    help="""\
A general utility for Fryhcs applications.

An application to load must be given with the '--app' option,
'FLASK_APP' environment variable, or with a 'wsgi.py' or 'app.py' file
in the current directory.
""",
)


def main():
    # 让python可以import .fry文件
    from fryhcs.fry.fryloader import install_path_hook
    install_path_hook()
    cli.main()


if __name__ == "__main__":
    main()



print("daisyui plugin")

def base_css():
    return {
        ":root, [data-theme=lofi]": {
            "color-scheme":        "light",
            "--rounded-box":       "0.25rem",
            "--rounded-btn":       "0.125rem",
            "--rounded-badge":     "0.125rem",
            "--animation-btn":     "0",
            "--animation-input":   "0",
            "--btn-focus-scale":   "1",
            "--tab-radius":        "0",
            "--cool":              "65 65 65",
            "--coll-content":      "250 250 250",
            "--coll-focus":        "30 30 30",
            "--base-200":          "40 40 40",
            "--base-300":          "30 30 30",
            "--base-content":      "100 100 100",
        },
        "@keyframes button-pop": {
          "0%": {
            "transform": "scale(var(--btn-focus-scale, 0.98))",
          },
          "40%": {
            "transform": "scale(1.02)",
          },
          "100%": {
            "transform": "scale(1)",
          },
        },
    }

def utilities():
    return [
        # 2023.12.16 暂不支持动态utility，不支持utility名中包含变量，semantic_color已经可以解决绝大部分变量的需求了
        ## 变量定义格式为<name:type>, :type可以省略，省略后默认是'DEFAULT'类型
        #('border-<color:state-color>', 
        #    # 在样式值中通过{color}引用变量的值，通过semantic-color['value']计算得出
        #    ('border-color', '{color}')),
        ('btn',
          ('@apply', 'gap-2 font-semibold no-underline border-base-200',
            'bg-base-200 text-base-content outline-base-content'),
          ('active:hover:&, active:focus:&', 
            ('animation', 'button-pop 0s ease-out'),
            ('transform', 'scale(var(--btn-focus-scale, 0.97))')),
          ('hover-hover:hover:&, active:&', ('@apply', 'border-base-300 bg-base-300')),
          ('focus-visible:&', ('@apply', 'outline outline-2 outline-offset-2')),
          ('border-width', 'var(--border-btn, 1px)'),
          ('animation', 'button-pop var(--animation-btn, 0.25s) ease-out'),
          ('text-transform', 'var(--btn-text-case, uppercase)')),

        # 考虑到性能，不支持在@apply的utility中引用变量（如下面的border-<color>）
        # utility中引用变量能极大简化插件编写，但会让fry转css的过程计算量增加非常多，
        # 因为插件加载过程中@apply的utility包含变量时无法转化为样式，只能在某个具体utility转
        # css时才能转样式，而这个utility所依赖的utility又会依赖其他utility，这个链条
        # 会非常长。
        # 不支持引用变量后，插件加载时其中定义的utility直接编译到样式，fry中的utility转css会加快很多。
        #    ('btn-<color:state-color>', 
        #      ('@apply', 'border-<color> bg-<color> text-<color>-content outline-<color>'),
        #      ('hover-hover:hover:&', 
        #        ('@apply', 'border-<color>-focus bg-<color>-focus'))),

        ('btn-cool',
          ('@apply', 'border-cool bg-cool text-cool-content outline-cool'),
          ('hover-hover:hover:&',
            ('@apply', 'border-cool-focus bg-cool-focus'))),
    ]

def colors():
    return {
        'cool': 'rgb(var(--cool))',
        'cool-content': 'rgb(var(--cool-content))',
        'cool-focus': 'rgb(var(--cool-focus))',
        'base-200': 'rgb(var(--base-200))',
        'base-300': 'rgb(var(--base-300))',
        'base-content': 'rgb(var(--base-content))',
    }



from fryhcs.utils import static_url, component_name
from fryhcs.config import fryconfig
from fryhcs.spec import is_valid_html_attribute
from fryhcs.css.style import CSS
import types

def escape(s):
    return s.replace('"', '\\"')


class RenderException(Exception):
    pass


def render_children(children, page):
    chs = []
    for ch in children:
        if isinstance(ch, list): #tuple和GeneratorType都已在Element.render第四步转化为list
            chs += render_children(ch, page)
        elif isinstance(ch, Element):
            chs.append(ch.render(page))
        else:
            chs.append(ch)
    return chs

def combine_style(style1, style2):
    if isinstance(style1, dict) and isinstance(style2, dict):
        result = {}
        result.update(style1)
        result.update(style2)
        return result
    elif isinstance(style1, dict) and isinstance(style2, str):
        result = ' '.join(f'{k}: {v};' for k,v in style1)
        return result + style2.strip()
    elif isinstance(style1, str) and isinstance(style2, dict):
        result = ' '.join(f'{k}: {v};' for k,v in style2)
        return result + style1.strip()
    elif isinstance(style1, str) and isinstance(style2, str):
        style1 = style1.strip()
        style2 = style2.strip()
        if len(style1) > 0 and style1[-1] != ';':
            style1 += ';'
        return style1 + style2


def convert_utilities(utilities):
    result = {}
    csses = [CSS(value=utility) for utility in utilities.split()]
    for css in csses:
        if not css.valid:
            raise RenderException(f"Invalid utility '{css.value}' in '@apply'")
        if css.wrappers or css.selector_template != css.default_selector_template:
            raise RenderException(f"Modifier is not allowed in '@apply': '{css.value}'")
        result.update(css.styles)
    return result


class ClientRef(object):
    def __init__(self, name):
        self.name = name
        self.component = 0

    def hook(self, component):
        if self.component == 0:
            self.component = component

    def __str__(self):
        return f'{self.name}-{self.component}'


class ClientEmbed(object):
    def __init__(self, embed_id):
        self.embed_id = embed_id
        self.component = 0

    def hook(self, component):
        if self.component == 0:
            self.component = component

    def __str__(self):
        if self.component == 0:
            return str(self.embed_id)
        else:
            return f'{self.component}/{self.embed_id}'

# 组件实例ID的html元素属性名
component_id_attr_name = 'data-fryid'
# js嵌入值的html元素属性名
client_embed_attr_name = 'data-fryembed'
# html元素引用的元素属性名，放在该html元素上
# 1. 通过ref引用的元素，生成'name-id'结构
# 2. 通过refall引用的元素列表，生成'name:a-id'
# 3. 对于html元素的引用，'name-id'中的id是引用自己的组件的id
# 4. 对于子组件元素的引用，放到父组件script的textContent中
client_ref_attr_name = 'data-fryref'

# 组件名的html元素属性名，只放在组件script上
component_name_attr_name = 'data-fryname'

# 组件js脚本url的html元素属性名，只放在组件script上
component_url_attr_name = 'data-fryurl'

# 组件元素作为一个模板时的无值属性名，组件元素用，不会最终生成到html中
component_template_attr_name = 'frytemplate'

# 组件元素生成<template>时添加的html元素属性名，值与组件实例id相同
component_template_id_atttr_name = 'data-frytid'

children_attr_name = 'children'
call_client_script_attr_name = 'call-client-script'
class_attr_name = 'class'
style_attr_name = 'style'
type_attr_name = 'type'
# 2023.11.30 py和js中的simple_quote也放入css utility检查范围，
#            $style和$class这种复杂处理方式不再需要
## 使用动态数据生成utility的属性名
#utility_attr_name = '$style'
# 引用属性名
ref_attr_name = 'ref'
# 引用列表属性名
refall_attr_name = 'refall'

class Element(object):

    def __init__(self, name, props=None, rendered=False):
        self.name = name
        self.props = {} if props is None else props
        self.rendered = rendered
        self.cid = 0

    def is_component(self):
        if self.rendered:
            return component_id_attr_name in self.props
        else:
            return callable(self.name) #inspect.isfunction(self.name) #or inspect.isclass(self.name)

    def get_style(self, style_name):
        if 'style' in self.props:
            return self.props['style'].get(style_name, None)
        utilities = []
        if not self.rendered:
            for k, v in self.props.items():
                if (v == '' or v is True) and not is_valid_html_attribute(self.name, k):
                    utilities.append(k)
        if 'class' in self.props:
            utilities += self.props['class'].split(' ')
        for utility in utilities:
            value = CSS(value=utility).get_style(style_name)
            if value:
                return value
        return None

    def tolist(self):
        def convert(v):
            if isinstance(v, (tuple, types.GeneratorType)):
                return list(v)
            else:
                return v
        def handle(v):
            if isinstance(v, Element):
                v.props = {key: convert(value) for key, value in v.props.items()}
                handle(v.props)
            elif isinstance(v, dict):
                for key, value in v.items():
                    value = convert(value)
                    v[key] = value
                    handle(value)
            elif isinstance(v, list):
                for i, value in enumerate(v):
                    value = convert(value)
                    v[i] = value
                    handle(value)
        handle(self)

    def hook_client_embed(self, component):
        def hook(v):
            if isinstance(v, (ClientEmbed, ClientRef)):
                v.hook(component)
            elif isinstance(v, list): #tuple和GeneratorType在render第四步都已转化为list
                for lv in v:
                    hook(lv)
            elif isinstance(v, dict):
                for lv in v.values():
                    hook(lv)
            elif isinstance(v, Element):
                hook(v.props)
        hook(self.props)

    def collect_client_embed(self, component):
        # 本函数中处理的component已经过渲染，所有element全是html element
        def collect(e):
            children = e.props.get(children_attr_name, [])
            for ch in children:
                if isinstance(ch, Element):
                    collect(ch)
            embeds = e.props.get(client_embed_attr_name, [])
            refs = e.props.get(client_ref_attr_name, [])
            for key in list(e.props.keys()):
                if key in (client_embed_attr_name, children_attr_name):
                    continue
                value = e.props.get(key)
                if isinstance(value, ClientEmbed) and value.component == component:
                    if e.name == 'script': # 暂时保留，已不支持
                        value.embed_id = f'{value.embed_id}-object-{key}'
                    elif key[0] == '@':
                        value.embed_id = f'{value.embed_id}-event-{key[1:]}'
                    elif key.startswith('$$'):
                        value.embed_id = f'{value.embed_id}-attr-{key[2:]}'
                    elif key == '*':
                        value.embed_id = f'{value.embed_id}-text'
                    else:
                        raise RenderException(f"Invalid client embed key '{key}' for element '{e.name}'")
                    embeds.append(value)
                    e.props.pop(key)
                elif isinstance(value, ClientRef) and value.component == component:
                    refs.append(value)
                    e.props.pop(key)
            if embeds:
                e.props[client_embed_attr_name] = embeds
            if refs:
                e.props[client_ref_attr_name] = refs
        collect(self)

    def render(self, page):
        """
        返回渲染后的元素。
        所有组件元素被渲染为基础元素（HTML元素），子元素列表中的子元素列表被摊平，属性值中不应再有元素
        """
        if self.rendered:
            return self

        if callable(self.name): #inspect.isfunction(self.name):
            # 渲染函数组件元素流程：
            # 1. 生成页面内组件实例对应的script元素，附加到页面后，
            #    得到组件实例唯一编号。
            #    组件函数每执行一次，返回该组件的一个实例。页面中
            #    每个组件实例都有一个页面内唯一编号。
            #    将组件名和组件实例ID附加到代表组件的script元素上，
            #    script是用来记录当前组件信息的，包括组件id，名字，
            #    以及后面可能的组件js参数
            component = {}
            cnumber = page.add_component(component)

            # 2. 预处理父组件传来的frytemplate/ref/refall/@event/class
            #    2.1 将本组件上定义的给父组件js脚本用的ref/refall记录到page
            #        上，在生成父组件的script元素时加到data-fryref上；
            #    2.2 将父组件传来的@event事件处理函数暂存，渲染完成后添加到
            #        本组件树的树根元素上;
            #    2.3 将父组件传来的class值暂存，渲染完成后附加到本组件树的数根元素上。
            peventhandlers = []
            pclass = ''
            istemplate = self.props.pop(component_template_attr_name, False)
            for key, value in list(self.props.items()):
                if isinstance(value, ClientRef):
                    self.props.pop(key)
                    pcid = value.component
                    if pcid == 0:
                        raise RuntimeError(f"Invalid ClientRef {value.name}")
                    name = 't:' + value.name if istemplate else value.name
                    page.add_ref(pcid, name, cnumber)
                elif key[0] == '@':
                    name = key[1:]
                    if istemplate:
                        raise RuntimeError(f"Can't attach event handler {name} to a frytemplate")
                    self.props.pop(key)
                    value.embed_id = f'{value.embed_id}-event-{name}'
                    peventhandlers.append(value)
                elif key == 'class':
                    self.props.pop(key)
                    pclass = value

            # 3. 执行组件函数，返回未渲染的原始组件元素树
            #    唯一不是合法python identifier的ref(:jsname)、refall(:jsname)和已经在上一步
            #    删除，此时self.props的key应该都是合法的python identifier，可以
            #    **self.props用来给函数调用传参。
            #    元素树中的js嵌入值以ClientEmbed对象表示，元素树中
            #    的ClientEmbed对象只能是新生成的本组件js嵌入值。
            #    本组件js嵌入值中(暂时)不带组件实例唯一编号，通过下一步hook_client_embed
            #    将组件实例唯一编号附加到js嵌入值中。
            #    其中：
            #    * 元素树中html元素属性和文本中的js嵌入值都被移到
            #      所在元素的data-fryembed属性值列表中；
            #    * 元素树中子组件元素属性中的js嵌入值只有ref和refall，已经在
            #      上一步中处理，所以子组件元素属性中不存在js嵌入值
            result = self.name(**self.props)
            if not isinstance(result, Element):
                raise RuntimeError(f"Function '{self.name.__name__}' should return Element")

            # 4. 转化Generator/tuple为list，Generator只能遍历一次，后面会有多次的遍历
            #    tuple无法改变内部数据，也变为list
            result.tolist()

            # 5. 将组件实例唯一编号挂载到组件元素树的所有本组件生成的
            #    js嵌入值上，使每个js嵌入值具有页面内唯一标识，
            #    标识格式为：组件实例唯一编号/js嵌入值在组件内唯一编号
            result.hook_client_embed(cnumber)

            # 6. 从原始组件元素树根元素的属性中取出calljs属性值
            calljs = result.props.pop(call_client_script_attr_name, False)

            # 7. 原始组件元素树渲染为最终的html元素树，
            element = result.render(page)
            element.cid = cnumber

            # 8. 此时已hook到组件实例的js嵌入值已挂载到html元素树上的合适
            #    位置，将这些js嵌入值收集到`client_embed_attr_name('data-fryembed')`
            #    和`client_ref_attr_name('data-fryref')`属性上
            element.collect_client_embed(cnumber)
            
            # 9. 将组件实例ID附加到组件html元素树树根元素的组件id列表'data-fryid'上
            cid = element.props.get(component_id_attr_name, '')
            element.props[component_id_attr_name] = f'{cnumber} {cid}' if cid else str(cnumber)

            # 10. 将父组件传来的事件处理函数记录到根元素上
            embeds = element.props.get(client_embed_attr_name, [])
            embeds += peventhandlers
            if embeds:
                element.props[client_embed_attr_name] = embeds

            # 11. 将父组件传来的class附加到根元素上
            selfclass = element.props.get(class_attr_name, '')
            if selfclass and pclass:
                selfclass += ' ' + pclass
            elif pclass:
                selfclass = pclass
            if selfclass:
                element.props[class_attr_name] = selfclass

            # 12. 将子组件实例的引用附加到script上(ref和refall都编码到refs中了)
            component['name'] = component_name(self.name)
            component['refs'] = page.child_refs(cnumber)

            # 13. 若当前组件存在js代码，记录组件与脚本关系，然后将组件js参数加到script脚本上
            if calljs:
                uuid, args = calljs
                component['url'] = f'{static_url(fryconfig.js_url)}{uuid}.js'
                component['args'] = {k:v for k,v in args}
                page.hasjs = True

            # 14. 对于组件模板，使用<template>包装起来
            if istemplate:
                props = {
                    component_template_id_atttr_name: cnumber,
                    children_attr_name: [element]
                }
                element = Element('template', props, True)
        elif isinstance(self.name, str):
            props = {}
            #style = {} 
            classes = []
            for k in list(self.props.keys()):
                v = self.props[k]
                if k == children_attr_name:
                    props[k] = render_children(v, page)
                elif isinstance(v, Element):
                    props[k] = v.render(page)
                #elif k == utility_attr_name:
                #    if isinstance(v, (list, tuple, types.GeneratorType)):
                #        v = ' '.join(v)
                #    elif not isinstance(v, str):
                #        raise RenderException(f"Invalid $style value: '{v}'")
                #    style = combine_style(style, convert_utilities(v))
                #elif k == style_attr_name:
                #    style = combine_style(style, v)
                elif is_valid_html_attribute(self.name, k):
                    props[k] = v
                elif v is True:
                    classes.append(k)
                elif isinstance(v, str):
                    values = v.split()
                    if not values:
                        values = ['']
                    classes.extend(CSS(k, value).to_class() for value in values) 
                else:
                    props[k] = v
            if classes:
                currclass = props.get(class_attr_name, '')
                classes = ' '.join(classes)
                if currclass:
                    currclass += ' ' + classes
                else:
                    currclass = classes
                props[class_attr_name] = currclass
            #if style:
            #    props[style_attr_name] = style
            element = Element(self.name, props, True)
        else:
            raise RenderException(f"invalid element name '{self.name}'")

        element.page = page
        return element


    def __str__(self):
        if not self.rendered:
            return '<Element(not rendered)>'

        children = self.props.pop(children_attr_name, None)
        attrs = []
        for k, v in self.props.items():
            if isinstance(v, dict):
                values = []
                for k1, v1 in v.items():
                    values.append(f"{k1}: {v1};")
                value = ' '.join(values)
            elif isinstance(v, (list, tuple, types.GeneratorType)):
                value = ' '.join(str(x) for x in v)
            elif v is True:
                value = ''
            elif v is False:
                continue
            else:
                value = str(v)
            if value:
                attrs.append(f'{k}="{escape(value)}"')
            else:
                attrs.append(k)
        if attrs:
            attrs = ' ' + ' '.join(attrs)
        else:
            attrs = ''
        if children is None:
            return f'<{self.name}{attrs} />'
        else:
            children = ''.join(str(ch) for ch in children)
            return f'<{self.name}{attrs}>{children}</{self.name}>'


Element.ClientEmbed = ClientEmbed
Element.ClientRef = ClientRef



from pathlib import Path
import os

from .style import CSS
from .collector import Collector
from .color import theme_color_styles
from fryhcs.config import fryconfig


class CSSGenerator():
    def __init__(self, input_files, output_file):
        self.input_files = input_files
        self.output_file = Path(output_file).absolute()
        self.reset()

    def reset(self):
        self.selectors = set()
        self.init_collector()

    def init_collector(self):
        self.collector = Collector()
        for file in self.input_files:
            if isinstance(file, str):
                self.collector.add_file(file)
            elif isinstance(file, (tuple, list)):
                dir = file[0]
                for glob in file[1:]:
                    self.collector.add_glob(dir, glob)

    def generate(self, input_file=None):
        """
        如果input_file为空，则是一个全量生成css；
        如果input_file不为空，则是一个增量生成css，只生成input_file的css，
        此时注意，input_file中的**所有**样式都会重新生成一遍，附加到css
        文件最后，会有很多样式重复生成，这是故意为之，目的是防止样式顺序
        错误，因为样式的顺序决定了样式的优先级，同样优先级的selector下，
        后面的样式会覆盖前面的样式。
        """
        # 1. collect all utilities from configed files
        if input_file:
            collector = Collector()
            collector.add_file(input_file)
            incremental = True
        else:
            self.reset()
            collector = self.collector
            incremental = False

        collector.collect_attrs()

        # 2. generate and write css
        self.output_file.parent.mkdir(parents=True, exist_ok=True)
        preflight = os.path.join(os.path.dirname(__file__), 'preflight.css')
        preflight = Path(preflight)
        with self.output_file.open('a' if incremental else 'w', encoding='utf-8') as f:
            if not incremental:
                with preflight.open('r', encoding='utf-8') as pf:
                    f.write(pf.read())
                f.write(theme_color_styles())
                from fryhcs.css.plugin import plugin_basecss
                basecss = plugin_basecss()
                f.write(basecss)
                preflight = fryconfig.config_root / 'preflight.css'
                if preflight.is_file():
                    with preflight.open('r', encoding='utf-8') as pf:
                        f.write(pf.read())
            csses = []
            for key, value in collector.all_attrs():
                css = CSS(key, value)
                if css.valid:
                    csses.append(css)
            for css in sorted(csses, key=lambda c: c.order):
                f.write(css.text())



from parsimonious import NodeVisitor, BadGrammar
from pathlib import Path
import sys
import re
import hashlib
from fryhcs.fry.grammar import grammar
from fryhcs.spec import is_valid_html_attribute
from fryhcs.css.style import CSS
from fryhcs.element import children_attr_name, call_client_script_attr_name, ref_attr_name, refall_attr_name

def escape(s):
    return s.replace('"', '\\"')

def quote_bracket_f_string(s):
    if not '"""' in s:
        quote = '"""'
    elif not "'''" in s:
        quote = "'''"
    else:
        raise BadGrammar(f"F-string '{s}' should not have both \"'''\" and '\"\"\"'")
    return f"f{quote}{s}{quote}"


class BaseGenerator(NodeVisitor):
    def __init__(self):
        self.client_embed_count = 0

    def inc_client_embed(self):
        count = self.client_embed_count
        self.client_embed_count = count + 1
        return count 

    def reset_client_embed(self):
        self.client_embed_count = 0

    def get_uuid(self, cname, node):
        sha1 = hashlib.sha1()
        sha1.update(node.text.encode('utf-8'))
        return f'{cname.lower()}-{sha1.hexdigest()}'

    def set_curr_file(self, file):
        self.curr_file = Path(file).absolute().resolve(strict=True)
        self.curr_dir  = self.curr_file.parent
        self.curr_root = None
        for p in self.curr_file.parents:
            init = p / '__init__.py'
            if not init.exists():
                self.curr_root = p
                break
        self.relative_dir = self.curr_dir.relative_to(self.curr_root)



#client_embed_attr_name = 'data-fryembed'

no_attr = 'no_attr'                   # ('no_attr', ...)
spread_attr = 'spread_attr'           # ('spread_attr', script): {script}
literal_attr = 'literal_attr'         # ('literal_attr', name, value): name="literal_value"
novalue_attr = 'novalue_attr'         # ('novalue_attr', name): name
py_attr = 'py_attr'                   # ('py_attr', name, pyscript): name={pyscript}
js_attr = 'js_attr'                   # ('js_attr', name, jscount): name=(jsscript)
#2023.11.24: 根据服务端由外而内，客户端由内而外的设计，不再需要jsop
#jsop_attr = 'jsop_attr'               # ('jsop_attr', name, pyscript): name=({pyscript})
#2023.10.27 不再支持元素作为属性值，参考语法文件fry.ppeg说明
#element_attr = 'element_attr'         # ('element_attr', name, element): name=<element></element>
#2023.11.16 不再支持python嵌入后跟js嵌入
#pyjs_attr = 'pyjs_attr'               # ('pyjs_attr', name, pyscript, jscount): name={pyscript}(jsscript)
#pyjsop_attr = 'pyjsop_attr'           # ('pyjsop_attr', name, pyscript, pyscript): name={pyscript1}({pyscript2})
fsjs_attr = 'fsjs_attr'               # ('fsjs_attr', name, value, jscount): name=[value](jsscript)
#fsjsop_attr = 'fsjsop_attr'           # ('fsjsop_attr', name, value, value): name=[value]({pyscript})
children_attr = 'children_attr'       # ('children_attr', [children])
jstext_attr = 'jstext_attr'           # ('jstext_attr', jscount)
#jsoptext_attr = 'jsoptext_attr'       # ('jsoptext_attr', value)
call_client_attr = 'call_client_attr' # ('call_client_attr', uuid, args)

def concat_kv(attrs):
    ats = []
    for attr in attrs:
        if isinstance(attr, (list, tuple)):
            atype = attr[0]
            if atype == spread_attr:
                ats.append(attr[1])
            elif atype == literal_attr:
                _, name, value = attr
                ats.append(f'"{name}": {value}')
            elif atype == novalue_attr:
                name = attr[1]
                ats.append(f'"{name}": True')
            elif atype == py_attr:
                _, name, value = attr
                ats.append(f'"{name}": {value}')
            elif atype == js_attr:
                _, name, jsvalue = attr
                # 对于ref/refall，只要保持名字与其他属性不冲突即可，前面加上
                # 一个冒号，如ref=(foo)，属性名即为":foo"，此名不会使用
                if name == ref_attr_name:
                    ats.append(f'":{jsvalue}": Element.ClientRef("{jsvalue}")')
                elif name == refall_attr_name:
                    ats.append(f'":{jsvalue}": Element.ClientRef("{jsvalue}:a")')
                else:
                    ats.append(f'"{name}": Element.ClientEmbed({jsvalue})')
            #elif atype == jsop_attr:
            #    _, name, value = attr
            #    ats.append(f'"{name}": {value}')
            #elif atype == element_attr:
            #    _, name, value = attr
            #    ats.append(f'"{name}": {value}')
            #elif atype == pyjs_attr:
            #    _, name, value, jscount = attr
            #    ats.append(f'"{name}": {value}')
            #    ats.append(f'"$${name}": Element.ClientEmbed({jscount})')
            #elif atype == pyjsop_attr:
            #    _, name, value, jsvalue = attr
            #    ats.append(f'"{name}": {value}')
            #    ats.append(f'"$${name}": {jsvalue}')
            elif atype == fsjs_attr:
                _, name, value, jscount = attr
                ats.append(f'"{name}": {value}')
                ats.append(f'"$${name}": Element.ClientEmbed({jscount})')
            #elif atype == fsjsop_attr:
            #    _, name, value, jsvalue = attr
            #    ats.append(f'"{name}": {value}')
            #    ats.append(f'"$${name}": {jsvalue}')
            elif atype == children_attr:
                ats.append(f'"{children_attr_name}": [{", ".join(attr[1])}]')
            elif atype == jstext_attr:
                _, jscount = attr
                ats.append(f'"*": Element.ClientEmbed({jscount})')
            #elif atype == jsoptext_attr:
            #    _, jsvalue = attr
            #    ats.append(f'"*": {jsvalue}')
            elif atype == call_client_attr:
                _, uuid, args = attr
                args = ', '.join(f'("{k}", {v})' for k,v in args)
                ats.append(f'"{call_client_script_attr_name}": ["{uuid}", [{args}]]')
            elif atype == no_attr:
                pass
            else:
                raise BadGrammar(f"Invalid attribute: {attr}")
        else:
            raise BadGrammar(f"Invalid attr: {attr}")
    return ats

# 检查并预处理html基本元素的属性
#
#   html基本元素是符合html规范的元素，支持无属性值属性、有属性值属性、事件处理器属性，属性有如下几种：
#   * `name`                               : 无值属性，如果是符合html规范的属性，正常传给浏览器引擎，否则放到class中
#                                            服务端：class中的值，或无值属性
#                                            浏览器：class中的值，或无值属性
#   * `name="literal_value"`               : 符合html规范的属性在客户端传给浏览器引擎，否则放到css中
#                                            服务端：class中的值，或正常html元素属性
#                                            浏览器：class中的值，或正常html元素属性
#   * `name='literal_value'`               : 符合html规范的属性在客户端传给浏览器引擎，否则放到css中
#                                            服务端：class中的值，或正常html元素属性
#                                            浏览器：class中的值，或正常html元素属性
#   * `name=(js_value)`                    : 这种格式目前只支持name为ref或refall，js_value是一个js变量名，用于将
#                                            当前元素赋值给一个js变量
#   * `@event=(js_handler)`                : 本组件的js事件处理函数
#                                            服务端：ClientEmbed对象
#                                            浏览器：data-fry-script一项
# 2023.11.24: 如下情况不再支持
# <del>
#   * `@event=({jsop_handler})`            : 父组件的js事件处理函数
#   * `@event={py_value}`                  : ClientEmbed类型的python值，父组件的事件处理函数
#                                            服务端：data-fry-script一项
#                                            浏览器：data-fry-script一项
# </del>
# 2023.11.30 py和js中的simple_quote也放入css utility检查范围，更加方便简单，
#            $style和$class这种复杂处理方式不再需要
# <del>
#   * `$name={py_value}`                   : python值在服务端渲染为常量赋值给属性name，传给浏览器引擎
#                                            目前支持的name只有"style"($style), 用于使用utility指定元素内置style，
#   * `$name='literal_value'`              : utility列表值在服务端转化为CSS，传给浏览器引擎
#   * `$name="literal_value"`              : utility列表值在服务端转化为CSS，传给浏览器引擎
#                                            目前支持的name只有"class"($class), 用于向CSSGenerator传递一些动态生成的、
#                                            运行期间才能看到的utility, 实际渲染中$class被忽略
#                                            注：$class不会出现在这里，在解析过程中就过滤掉了(see `visit_fry_kv_attribute`)
# </del>
#   * `name={py_value}`                    : python值在服务端渲染为常量传给浏览器引擎，不可以为ClientEmbed
#   * `{name}`                             : `name={name}`的简写
#                                            服务端：`name=py_value`，python数据值
#                                            浏览器：`name="py_value"`，字符串值，如果是ClientEmbed时生成data-fry-script一项
# 2023.11.16: 如下两种情况不再支持：
# <del>
#   * `name={py_value}(js_value)`          : python值，客户端js修改
#   * `name={py_value}({jsop_value})`      : python值，客户端父组件js修改
#                                            服务端：`name=py_value`，python数据值
#                                            浏览器：`name="py_value"`，字符串值，同时新增data-fry-script一项
# </del>
#   * `name=[literal_value](js_value)`     : 常量字符串，客户端js在水合后以js_value将其修改
# 2023.11.24: 如下不再支持
# <del>
#   * `name=[literal_value]({jsop_value})` : 常量字符串，客户端父组件js修改
#                                            服务端：`name=literal_value`，python数据值
#                                            浏览器：`name="literal_value"`，字符串值，同时新增data-fry-script一项
# </del>
#   * `{*python_list}`                     : python列表值，服务端渲染为常量传给浏览器引擎
#   * `{**python_dict}`                    : python字典值，服务端渲染为常量传给浏览器引擎
def check_html_element(name, attrs):
    for attr in attrs:
        atype = attr[0]
        if atype not in (novalue_attr, literal_attr, js_attr, py_attr, fsjs_attr, spread_attr):
            raise BadGrammar(f"Invalid attribute type '{atype}' in html element '{name}'")
        # 检查事件处理器
        if attr[1][0] == '@' and atype not in (js_attr, ):
            raise BadGrammar(f"Invalid attribute type '{atype}' for event handler '{attr[1]}' in html element '{name}'")

        if atype in (novalue_attr, literal_attr, py_attr):
            if attr[1] == 'frytemplate':
                raise BadGrammar('Only component element can have frytemplate attribute')

        # 2023.11.30: 不再支持$style
        ## 检查$style
        #if attr[1][0] == '$':
        #    if attr[1] not in ('$style',):
        #        raise BadGrammar("unsupported attribute name: '{attr[1]}'")
        #    if attr[1] == '$style' and atype != py_attr:
        #        raise BadGrammar(f"invalid attribute type '{atype}' for '$style' in html element '{name}': '{py_attr}' needed.")
        if atype == js_attr:
            if attr[1][0] == '@':
                attr[0] = py_attr
                attr[2] = f'Element.ClientEmbed({attr[2]})'
            elif attr[1] == ref_attr_name:
                attr[0] = py_attr
                attr[1] = f':{attr[2]}'
                attr[2] = f'Element.ClientRef("{attr[2]}")'
            elif attr[1] == refall_attr_name:
                attr[0] = py_attr
                attr[1] = f':{attr[2]}'
                attr[2] = f'Element.ClientRef("{attr[2]}:a")'
            else:
                raise BadGrammar(f"js_attr type can only be specified for event handler or ref/refall, not '{attr[1]}'")


# 检查并预处理组件元素的属性
#
#   组件元素以大写字母开头的名字作为元素名，元素名代表一个组件函数。
#   组件元素的属性作为python参数列表传给组件函数，所以组件元素只支持如下几种格式的属性：
#   * `name`                : 无值属性，等同于`name={name}`
#   * `name="literal_value"`: 常量字符串在服务端传给子组件
#                             服务端：常量字符串
#                             浏览器：不可见
#   * `name='literal_value'`: 常量字符串在服务端传给子组件
#                             服务端：常量字符串
#                             浏览器：不可见
#   * `name={py_value}`     : python值在服务端运行时传给子组件，可以是各种类型数据，但不包括ClientEmbed
#   * `{name}`              : `name={name}`的简写
#                             服务端：python数据
#                             浏览器：不可见
#   * `name=(js_value)`     : 目前只支持name为ref，此时js_value为父组件js代码的入参变量名,
#                             值为子组件的js对象。
#   * `@event=(js_handler)` : 传给子组件的js事件处理函数
# 2023.11.24: 如下情况不再支持
# <del>
#   * `name=({jsop_value})` : 父组件js值在客户端运行时传给子组件
#                             服务端：ClientEmbed对象
#                             浏览器：不可见
# </del>
#   * `{*python_list}`      : python列表值
#                             服务端：传递给python组件函数的props参数的一部分
#                             浏览器：不可见
#   * `{**python_dict}`     : python字典值
#                             服务端：传递给python组件函数的props参数的一部分
#                             浏览器：不可见
def check_component_element(name, attrs):
    for attr in attrs:
        atype = attr[0]
        if atype not in (novalue_attr, spread_attr, literal_attr, py_attr, js_attr):
            raise BadGrammar(f"Invalid attr '{atype}': Component element can only have novalue_attr, spread_attr, literal_attr, py_attr, js_attr")
        if atype in (literal_attr, py_attr) and not attr[1].isidentifier():
            raise BadGrammar(f"Invalid attibute name '{attr[1]}' on Component element '{name}', identifier needed.")
        if atype == novalue_attr:
            # 组件元素的无值参数等同于值为True变量 name => name={True}
            attr[0] = py_attr
            attr.append('True')
        elif atype == js_attr:
            if attr[1][0] == '@':
                attr[0] = py_attr
                attr[2] = f'Element.ClientEmbed({attr[2]})'
            elif attr[1] == ref_attr_name:
                attr[0] = py_attr
                attr[1] = f':{attr[2]}'
                attr[2] = f'Element.ClientRef("{attr[2]}")'
            elif attr[1] == refall_attr_name:
                attr[0] = py_attr
                attr[1] = f':{attr[2]}'
                attr[2] = f'Element.ClientRef("{attr[2]}:a")'
            else:
                raise BadGrammar(f"Only ref/refall/@event of component element can have js_attr, not '{attr[1]}'.")

class PyGenerator(BaseGenerator):
    def generate(self, tree):
        self.web_component_script = False
        self.client_script_args = {}
        self.refs = set()
        self.refalls = set()
        self.reset_client_embed()
        return self.visit(tree)

    def generic_visit(self, node, children):
        return children or node

    def visit_fry_script(self, node, children):
        return ''.join(str(ch) for ch in children)

    def visit_inner_fry_script(self, node, children):
        return ''.join(str(ch) for ch in children)

    def visit_fry_script_item(self, node, children):
        item = children[0]
        if isinstance(item, tuple):
            if item[0] == 'element':
                return item[1]
            else:
                raise BadGrammar
        return item 

    def visit_inner_fry_script_item(self, node, children):
        item = children[0]
        if isinstance(item, tuple):
            if item[0] == 'element':
                return item[1]
            else:
                raise BadGrammar
        return item 

    def visit_py_comment(self, node, children):
        return node.text

    def visit_inner_fry_brace(self, node, children):
        _, script, _ = children
        # inner_brace是正常的python脚本，需要原样输出
        return '{' + script + '}'

    def visit_fry_embed(self, node, children):
        _, script, _ = children
        # embed都是赋值表达式，可以直接加上小括号
        # 2024.10.22：作为fry_child时，不支持{*mylist}这种用法，直接{mylist}
        return ('fry_embed', '(' + script.strip() + ')')

    def visit_triple_single_quote(self, node, children):
        return node.text

    def visit_triple_double_quote(self, node, children):
        return node.text

    def visit_single_quote(self, node, children):
        return node.text

    def visit_double_quote(self, node, children):
        return node.text

    def visit_py_simple_quote(self, node, children):
        return children[0]

    def visit_js_simple_quote(self, node, children):
        return children[0]

    def visit_less_than_char(self, node, children):
        return '<'

    def visit_no_component_d_char(self, node, children):
        return 'd'

    def visit_py_normal_code(self, node, children):
        return node.text

    def visit_inner_py_normal_code(self, node, children):
        return node.text

    def visit_fry_component(self, node, children):
        cname, fryscript, template, _jsscript = children
        _type, name, attrs = template
        if self.web_component_script:
            uuid = self.get_uuid(cname, node)
            args = [(k,v) for k,v in self.client_script_args.items()]
            attrs.insert(0, [call_client_attr, f'{(self.relative_dir/uuid).as_posix()}', args])
        self.web_component_script = False
        self.client_script_args = {}
        self.refs = set()
        self.refalls = set()
        self.reset_client_embed()
        attrs = concat_kv(attrs)
        return f'def {cname}{fryscript}return Element({name}, {{{", ".join(attrs)}}})'

    def visit_fry_component_header(self, node, children):
        _def, _, cname, _ = children
        return cname

    def visit_fry_component_name(self, node, children):
        return node.text

    def visit_fry_web_template(self, node, children):
        _l, _, element, _, _r = children
        return element

    def visit_fry_root_element(self, node, children):
        name, attrs = children[0]
        if name == 'script':
            raise BadGrammar("'script' can't be used as the normal element name")
        return ('rootelement', name, attrs)

    def visit_fry_element(self, node, children):
        name, attrs = children[0]
        attrs = concat_kv(attrs)
        return ('element', f'Element({name}, {{{", ".join(attrs)}}})')

    def visit_fry_fragment(self, node, children):
        _, fry_children, _ = children
        return ('"div"', [[children_attr, fry_children]])

    def visit_fry_self_closing_element(self, node, children):
        _, name, attrs, _, _ = children
        if not name:
            raise BadGrammar
        if name[0].islower():
            check_html_element(name, attrs)
            name = f'"{name}"'
        else:
            check_component_element(name, attrs)
        # self closing元素不能有children属性，否则组件函数的参数会出错。
        #attrs.append([children_attr,[]])
        return (name, attrs)

    def visit_fry_void_element(self, node, children):
        _, name, attrs, _, _ = children
        if not name:
            raise BadGrammar
        # void element都是html元素
        check_html_element(name, attrs)
        name = f'"{name}"'
        return (name, attrs)

    def visit_fry_paired_element(self, node, children):
        start, fry_children, end = children
        start_name, attrs = start
        end_name = end
        if start_name != end_name:
            raise BadGrammar(f'start_name "{start_name}" is not the same with end_name "{end_name}"')
        name = start_name
        if not name:
            raise BadGrammar
        elif name[0].islower():
            check_html_element(name, attrs)
            name = f'"{name}"'
        else:
            check_component_element(name, attrs)

        attrs.append([children_attr, fry_children])

        return (name, attrs)

    def visit_fry_start_tag(self, node, children):
        _, start_name, attrs, _, _ = children
        return start_name, attrs

    def visit_fry_end_tag(self, node, children):
        _, name, _, _ = children
        return name

    def visit_fry_element_name(self, node, children):
        return node.text

    def visit_fry_void_element_name(self, node, children):
        return node.text

    def visit_space(self, node, children):
        return node.text

    def visit_maybe_space(self, node, children):
        return node.text

    def visit_fry_attributes(self, node, children):
        # 过滤掉空属性
        return [ch for ch in children if ch]

    def visit_fry_spaced_attribute(self, node, children):
        space, attr = children
        if not space and attr:
            # 元素属性前最好有空格
            print(f"Warning: attribute {attr} should be prefixed with white space.")
        return attr

    def visit_fry_attribute(self, node, children):
        return children[0]

    def visit_same_name_attribute(self, node, children):
        _l, _, identifier, _, _r = children
        return [py_attr, identifier, identifier]

    def visit_py_identifier(self, node, children):
        return node.text

    def visit_fry_embed_spread_attribute(self, node, children):
        _lbrace, _, stars, _, script, _rbrace = children
        if stars.text == '*':
            return [spread_attr, "**{ key: True for key in (" + script + ")}"]
        return [spread_attr, '**(' + script + ')']

    #def visit_fry_client_embed_attribute(self, node, children):
    #    value, _, _css_literal = children
    #    _name, literal, client_embed = value
    #    kvs = [(name, '""') for name in literal.split()]
    #    count = self.inc_client_embed()
    #    return (client_embed_attr_name, kvs, str(count))

    #def visit_fry_event_attribute(self, node, children):
    #    _at, _identifier, _, _equal, _, _client_embed = children
    #    count = self.inc_client_embed()
    #    return (client_embed_attr_name, [], str(count))

    def visit_fry_kv_attribute(self, node, children):
        name, _, _, _, value = children
        # 2023.11.30: 不再支持$class和$style，将simple_quote识别为utility
        #if name == '$class':
        #    if not isinstance(value, str):
        #        raise BadGrammar("'$class' can only have literal value")
        #    return None 
        if isinstance(value, str):
            return [literal_attr, name, value]
        elif isinstance(value, tuple):
            if value[0] == 'joint_embed':
                _, quoted_literal, client_embed = value
                if client_embed[0] == 'js_embed':
                    count = self.inc_client_embed()
                    return [fsjs_attr, name, quoted_literal, str(count)]
                else:
                    raise BadGrammar
            #elif value[0] == 'fry_js_embed':
            #    _, embed, client_embed = value
            #    if client_embed[0] == 'local_js_embed':
            #        count = self.inc_client_embed()
            #        return [pyjs_attr, name, embed, str(count)]
            #    elif client_embed[0] == 'jsop_embed':
            #        return [pyjsop_attr, name, embed, client_embed[1]]
            #    else:
            #        raise BadGrammar
            elif value[0] == 'fry_embed':
                _, embed = value
                return [py_attr, name, embed]
            elif value[0] == 'js_embed':
                if name in (ref_attr_name, refall_attr_name):
                    isall = name == refall_attr_name
                    v = value[1].strip()
                    if not v.isidentifier():
                        raise BadGrammar(f"Ref name '{v}' is not a valid identifier")
                    if isall:
                        if v in self.refs:
                            raise BadGrammar(f"Ref name '{v}' exists, please use another name for 'refall'")
                        self.refalls.add(v)
                    else:
                        if v in self.refs or v in self.refalls:
                            raise BadGrammar(f"Duplicated ref name '{v}', please use 'refall'")
                        self.refs.add(v)
                else:
                    v = self.inc_client_embed()
                return [js_attr, name, str(v)]
            #elif value[0] == 'jsop_embed':
            #    return [jsop_attr, name, value[1]]
            #elif value[0] == 'element':
            #    return [element_attr, name, value[1]]
            else:
                raise BadGrammar(f'Invalid attribute value: {value[0]}')
        else:
            raise BadGrammar(f'Invalid attribute value: {value}')

    def visit_fry_novalue_attribute(self, node, children):
        name, _ = children
        return [novalue_attr, name]

    def visit_fry_attribute_name(self, node, children):
        return node.text

    def visit_fry_attribute_value(self, node, children):
        return children[0]

    #def visit_fry_attr_value_embed(self, node, children):
    #    embed, _, client_embed, _, _css_literal = children
    #    return ('embed_value', embed, client_embed)

    #def visit_fry_attr_value_client_embed(self, node, children):
    #    value, _, _css_literal = children
    #    return value #('client_embed', literal, client_embed)

    #def visit_fry_css_literal(self, node, children):
    #    _colon, _, value = children
    #    return value

    #def visit_maybe_css_literal(self, node, children):
    #    if not children:
    #        return ''
    #    return children[0]

    def visit_fry_children(self, node, children):
        return [ch for ch in children if ch]

    def visit_fry_child(self, node, children):
        frychild = children[0]
        if isinstance(frychild, str):
            return frychild
        elif isinstance(frychild, tuple):
            #if frychild[0] == 'fry_js_embed':
            #    _, embed, client_embed = frychild
            #    if client_embed[0] == 'local_js_embed':
            #        attr = jstext_attr
            #        value = str(self.inc_client_embed())
            #    elif client_embed[0] == 'jsop_embed':
            #        attr = jsoptext_attr
            #        value = client_embed[1]
            #    else:
            #        raise BadGrammar
            #    attrs = [[attr, value],
            #             [children_attr, [embed]]]
            #    attrs = concat_kv(attrs)
            #    return f'Element("span", {{{", ".join(attrs)}}})'
            #elif frychild[0] == 'fry_embed':
            if frychild[0] == 'fry_embed':
                _, embed = frychild
                # 2024.10.12: 不支持{*mylist}这种语法，会导致python语法错误：
                # SyntaxError: cannot use starred expression here
                if embed[0] == '(' and embed[1] == '*':
                    raise BadGrammar('{' + embed[1:-1] + '} is not allowed in component template')
                return embed
            elif frychild[0] == 'joint_embed':
                _, quoted_literal, client_embed = frychild
                if client_embed[0] == 'js_embed':
                    attr = jstext_attr
                    value = str(self.inc_client_embed())
                #elif client_embed[0] == 'jsop_embed':
                #    attr = jsoptext_attr
                #    value = client_embed[1]
                else:
                    raise BadGrammar
                attrs = [[attr, value],
                         [children_attr, [quoted_literal]]]
                attrs = concat_kv(attrs)
                return f'Element("span", {{{", ".join(attrs)}}})'
            elif frychild[0] == 'element':
                return frychild[1]
        else:
            raise BadGrammar(f'Invalid fry_child "{frychild}"')

    #def visit_fry_js_embed(self, node, children):
    #    fry_embed, _, js_embed = children
    #    _name, fry = fry_embed
    #    return ('fry_js_embed', fry, js_embed)

    def visit_joint_embed(self, node, children):
        quoted_literal, _, js_embed = children
        return ('joint_embed', quoted_literal, js_embed)

    def visit_bracket_f_string(self, node, children):
        _l, body, _r = children
        return quote_bracket_f_string(body)

    def visit_bracket_f_string_body(self, node, children):
        return node.text

    def visit_single_f_string(self, node, children):
        if '{' in node.text:
            return 'f' + node.text
        else:
            return node.text

    def visit_double_f_string(self, node, children):
        if '{' in node.text:
            return 'f' + node.text
        else:
            return node.text

    def visit_fry_text(self, node, children):
        value = re.sub(r'(\s+)', lambda m: ' ', node.text).strip()
        if not value or value == ' ':
            return ''
        return f'"{escape(value)}"'

    def visit_no_embed_char(self, node, children):
        return node.text

    # 脚本元素的元素名为script，代表了一个组件对应的js脚本，一个组件最多有一个脚本元素。
    # 脚本元素的属性作为js参数列表传给脚本代码，并且脚本代码需要在编译期生成，属性名需要在编译期可见，
    # 不能依赖python运行期的信息，所以脚本元素只支持如下几种格式的属性：
    # * `name`                : 无值属性，用于定义一个与python变量同名的js局部变量，相当于 name={name}
    # * `name="literal_value"`: 常量字符串在客户端传给js脚本
    #                           服务端：`data-name="literal_value"`
    #                           浏览器：`data-name="literal_value"`
    # * `name='literal_value'`: 常量字符串在客户端传给js脚本
    #                           服务端：`data-name='literal_value'`
    #                           浏览器：`data-name="literal_value"`
    # * `name={py_value}`     : python值作为字符串在客户端运行时传给js脚本，在客户端是一个常量字符串
    # * `{name}`              : `name={name}`的简写
    #                           服务端：`data-name=py_value`，python数据值
    #                           浏览器：`data-name="py_value"`，字符串值。
    # 2023.11.24: 根据服务端由外而内，客户端由内而外的设计，不再需要jsop
    # <del>
    # * `name=({py_value})`   : ClientEmbed值在客户端运行时传给本组件js脚本，是父组件的js值
    #                           服务端：`data-fryembed=[ClientEmbed]`，ClientEmbed值
    #                           浏览器：`data-fryembed="4/3-object-foo"`，父组件js值
    # </del>
    # name不能以'fry'开头
    def visit_web_script(self, node, children):
        self.web_component_script = True
        _, _begin, attributes, _, _gt, _script, _end = children
        for attr in attributes:
            # TODO spread_attr是不是也可以用在这里？
            if attr[0] not in (novalue_attr, literal_attr, py_attr):
                raise BadGrammar("script attributes can only be novalue_attr, literal_attr or py_attr")
            name = attr[1]
            if not name.isidentifier():
                raise BadGrammar(f"Script argument name '{name}' is not valid identifier.")
            if name in self.client_script_args:
                raise BadGrammar(f"Script argument name duplicated: '{name}'.")

            if name in self.refs or name in self.refalls:
                raise BadGrammar(f"Script argument name '{name}' duplicated with ref/refall names")
            if attr[0] == novalue_attr:
                # 所有无值变量都是值为True的变量
                # 将这些attr转化为py_attr: foo ==> foo={True}
                attr.append('True')
                attr[0] = py_attr
            atype, k, v = attr
            if k.startswith('fry'):
                raise BadGrammar(f"Prefix 'fry' is reserved, <script> attribute name can't be started with 'fry'")
            self.client_script_args[k] = v
        return ''

    def visit_html_comment(self, node, children):
        return ''

    def visit_js_embed(self, node, children):
        self.web_component_script = True
        # 返回js script内容，在ref DOM元素时有用
        return ('js_embed', node.text[1:-1])

    #def visit_jsop_embed(self, node, children):
    #    _l, script, _r = children
    #    return ('jsop_embed', script)


def fry_to_py(source, path):
    """
    fry文件内容转成py文件内容
    """
    tree = grammar.parse(source)
    generator = PyGenerator()
    generator.set_curr_file(path)
    return generator.generate(tree)



from django.conf import settings
from django.urls import path

from .views import component, check_hotreload

app_name = 'fryhcs'

urlpatterns = [
#    path('component', components, name="component"),
]

if settings.DEBUG:
    urlpatterns += [
        path('_check_hotreload', check_hotreload, name="check_hotreload"),
    ]



from fryhcs.element import Element, component_id_attr_name, component_name_attr_name, component_url_attr_name, children_attr_name, type_attr_name
from fryhcs.utils import static_url
from fryhcs.config import fryconfig
from importlib import import_module
import json

class Page(object):
    def __init__(self):
        # 记录当前已经处理的组件script元素列表，列表长度是当前正在处理的组件的ID
        self.components = {}
        # 组件实例ID到子组件引用/引用列表的映射关系, cid -> (refname -> childcid | list of childcid)
        self.cid2childrefs = {}
        self.hasjs = False

    def add_component(self, component):
        cid = len(self.components) + 1
        self.components[cid] = component
        component['cid'] = cid
        self.cid2childrefs[cid] = {} 
        return cid

    def add_ref(self, cid, refname, childcid):
        childrefs = self.cid2childrefs[cid]
        if refname in childrefs:
            refs = childrefs[refname]
        else:
            refs = set()
            childrefs[refname] = refs
        refs.add(childcid)

    def child_refs(self, cid):
        origin = self.cid2childrefs.get(cid, {})
        refs = {}
        for name, ids in origin.items():
            if name.endswith(':a'):
                refs[name[:-2]] = sorted(ids)
            else:
                if len(ids) != 1:
                    raise RuntimeError(f"More than ONE ref value for '{name}'.")
                refs[name] = ids.pop()
        return refs


def render(element, **kwargs):
    page = Page()
    if isinstance(element, Element):
        element = element.render(page)
    elif callable(element) and getattr(element, '__name__', 'anonym')[0].isupper():
        element = Element(element, kwargs).render(page)
    elif isinstance(element, str):
        if '.' in element:
            module, _, comp = element.rpartition('.')
            module = import_module(module)
            component = getattr(module, comp)
            if callable(component) and comp[0].isupper():
                element = Element(component, kwargs).render(page)
        elif element and element[0].islower():
            element = Element(element, kwargs).render(page)
    return element


def html(content='div',
         args={},
         title='',
         lang='en',
         rootclass='',
         charset='utf-8',
         viewport="width=device-width, initial-scale=1.0",
         metas={},
         properties={},
         equivs={},
         autoreload=True,
        ):
    sep = '\n    '
    main_content = render(content, **args)
    page = main_content.page
    if main_content.name == 'body':
        body = main_content
    else:
        body = Element('body', dict(children=[main_content]), True)
    components = []
    for c in page.components.values():
        scriptprops = {
            type_attr_name: 'text/x-frydata',
            component_id_attr_name: c['cid'],
            component_name_attr_name: c['name'],
        }
        content = {}
        if 'url' in c:
            scriptprops[component_url_attr_name] = c['url']
            content['refs'] = c['refs']
            content['args'] = c['args']
        scriptprops[children_attr_name] = [json.dumps(content)]
        comp = Element('script', scriptprops, True)
        components.append(comp)
    if components:
        components = Element('div', dict(style=dict(display='none'), children=components), True)
        body.props[children_attr_name].append(components)
    if page.hasjs:
        # 此时必定存在script[data-fryurl]，所以js代码中无需校验script是否存在
        script = """
      const script = document.querySelector('script[data-fryurl]');
      const { hydrate } = await import(script.dataset.fryurl);
      await hydrate(document.documentElement);
"""
        hydrate_script = Element('script', dict(type='module', children=[script]), True)
        body.props[children_attr_name].append(hydrate_script)

    if fryconfig.debug and autoreload:
        script = f"""
      let serverId = null;
      let eventSource = null;
      let timeoutId = null;
      function checkAutoReload() {{
          if (timeoutId !== null) clearTimeout(timeoutId);
          timeoutId = setTimeout(checkAutoReload, 1000);
          if (eventSource !== null) eventSource.close();
          eventSource = new EventSource("{fryconfig.check_reload_url}");
          eventSource.addEventListener('open', () => {{
              console.log(new Date(), "Auto reload connected.");
              if (timeoutId !== null) clearTimeout(timeoutId);
              timeoutId = setTimeout(checkAutoReload, 1000);
          }});
          eventSource.addEventListener('message', (event) => {{
              const data = JSON.parse(event.data);
              if (serverId === null) {{
                  serverId = data.serverId;
              }} else if (serverId !== data.serverId) {{
                  if (eventSource !== null) eventSource.close();
                  if (timeoutId !== null) clearTimeout(timeoutId);
                  location.reload();
                  return;
              }}
              if (timeoutId !== null) clearTimeout(timeoutId);
              timeoutId = setTimeout(checkAutoReload, 1000);
          }});
      }}
      checkAutoReload();
"""
        autoreload = Element('script', dict(type='module', children=[script]), True)
        body.props[children_attr_name].append(autoreload)

    metas = sep.join(f'<meta name="{name}" content="{value}">'
                       for name, value in metas.items())
    properties = sep.join(f'<meta property="{property}" content="{value}">'
                            for property, value in properties.items())
    equivs = sep.join(f'<meta http-equiv="{equiv}" content="{value}">'
                            for equiv, value in equivs.items())
    # no need to use importmap
    #importmap = f'''
    #<script type="importmap">
    #  {{
    #    "imports": {{
    #      "fryhcs": "{static_url('js/fryhcs.js')}"
    #    }}
    #  }}
    #</script>
    #'''

    if rootclass:
        rootclass = f' class="{rootclass}"'
    else:
        rootclass = ''

    return f'''\
<!DOCTYPE html>
<html lang={lang}{rootclass}>
  <head>
    <meta charset="{charset}">
    <title>{title}</title>
    <meta name="viewport" content="{viewport}">
    {metas}
    {properties}
    {equivs}
    <link rel="stylesheet" href="{static_url(fryconfig.css_url)}">
  </head>
  {body}
</html>
'''
